from ..base import OptionsGroup, ParametrizedValue
from ..utils import listify


class Logger(ParametrizedValue):

    args_joiner = ','

    def __init__(self, alias, *args):
        self.alias = alias or ''
        super(Logger, self).__init__(*args)


class LoggerFile(Logger):
    """Allows logging into files."""

    name = 'file'
    plugin = 'logfile'

    def __init__(self, alias, filepath):
        """
        :param str|unicode alias: Logger alias.

        :param str|unicode filepath: File path.

        """
        super(LoggerFile, self).__init__(alias, filepath)


class LoggerSocket(Logger):
    """Allows logging into UNIX and UDP sockets."""

    name = 'socket'
    plugin = 'logsocket'

    def __init__(self, alias, addr_or_path):
        """
        :param str|unicode alias: Logger alias.

        :param str|unicode addr_or_path: Remote address or filepath.

            Examples:
                * /tmp/uwsgi.logsock
                * 192.168.173.19:5050

        """
        super(LoggerSocket, self).__init__(alias, addr_or_path)


class LoggerSyslog(Logger):
    """Allows logging into Unix standard syslog or a remote syslog."""

    name = 'syslog'
    plugin = 'syslog'

    def __init__(self, alias, app_name=None, facility=None, host=None):
        """
        :param str|unicode alias: Logger alias.

        :param str|unicode app_name:

        :param str|unicode facility:

        :param str|unicode host: Host and port for remote syslog.

        """
        args = []

        if host:
            self.name = 'rsyslog'
            self.requires_plugin = 'rsyslog'
            args.append(host)

        args.extend([app_name, facility])

        super(LoggerSyslog, self).__init__(alias, *args)


class LoggerRedis(Logger):
    """Allows logging into Redis.

    .. note:: Consider using ``dedicate_thread`` param.

    """

    name = 'redislog'
    plugin = 'redislog'

    def __init__(self, alias, host=None, command=None, prefix=None):
        """
        :param str|unicode alias: Logger alias.

        :param str|unicode host: Default: 127.0.0.1:6379

        :param str|unicode command: Command to be used. Default: publish uwsgi

            Examples:
                * publish foobar
                * rpush foo

        :param str|unicode prefix: Default: <empty>

        """
        super(LoggerRedis, self).__init__(alias, host, command, prefix)


class LoggerMongo(Logger):
    """Allows logging into Mongo DB.

    .. note:: Consider using ``dedicate_thread`` param.

    """

    name = 'mongodblog'
    plugin = 'mongodblog'

    def __init__(self, alias, host=None, collection=None, node=None):
        """
        :param str|unicode alias: Logger alias.

        :param str|unicode host: Default: 127.0.0.1:27017

        :param str|unicode collection: Command to be used. Default: uwsgi.logs

        :param str|unicode node: An identification string for the instance
            sending logs Default: <server hostname>

        """
        super(LoggerMongo, self).__init__(alias, host, collection, node)


class LoggerZeroMq(Logger):
    """Allows logging into ZeroMQ sockets."""

    name = 'zeromq'
    plugin = 'logzmq'

    def __init__(self, alias, connection_str):
        """
        :param str|unicode alias: Logger alias.

        :param str|unicode connection_str:

            Examples:
                * tcp://192.168.173.18:9191

        """
        super(LoggerZeroMq, self).__init__(alias, connection_str)


class Encoder(ParametrizedValue):

    name_separator = ' '


class EncoderPrefix(Encoder):
    """Add a raw prefix to each log msg."""

    name = 'prefix'

    def __init__(self, value):
        """

        :param str|unicode value: Value to be used as affix

        """
        super(EncoderPrefix, self).__init__(value)


class EncoderSuffix(EncoderPrefix):
    """Add a raw suffix to each log msg"""

    name = 'suffix'


class EncoderNewline(Encoder):
    """Add a newline char to each log msg."""

    name = 'nl'


class EncoderGzip(Encoder):
    """Compress each msg with gzip (requires zlib)."""

    name = 'gzip'


class EncoderCompress(Encoder):
    """Compress each msg with zlib compress (requires zlib)."""

    name = 'compress'


class EncoderFormat(Encoder):
    """Apply the specified format to each log msg."""

    name = 'format'

    def __init__(self, template):
        """

        :param str|unicode template: Template string.
            Available variables are listed in ``FormatEncoder.Vars``.

        """
        super(EncoderFormat, self).__init__(template)

    class vars(object):
        """Variables available to use."""

        MESSAGE = '${msg}'
        """Raw log message (newline stripped)."""

        MESSAGE_NEWLINE = '${msgnl}'
        """Raw log message (with newline)."""

        TIME = '${unix}'
        """Current unix time."""

        TIME_MS = '${micros}'
        """Current unix time in microseconds."""

        # todo consider adding ${strftime:xxx} - strftime using the xxx format


class EncoderJson(EncoderFormat):
    """Apply the specified format to each log msg with each variable json escaped."""

    name = 'json'


class Logging(OptionsGroup):
    """Logging.

    * http://uwsgi.readthedocs.io/en/latest/Locks.html
    * http://uwsgi-docs.readthedocs.io/en/latest/LogFormat.html

    """

    class loggers(object):
        """Loggers available for ``add_logger()``."""

        file = LoggerFile
        mongo = LoggerMongo
        redis = LoggerRedis
        socket = LoggerSocket
        syslog = LoggerSyslog
        zeromq = LoggerZeroMq

        # todo consider adding other loggers: crypto, graylog2, systemd

    class encoders(object):
        """Loggers available for ``add_logger_encoder()``."""

        compress = EncoderCompress
        format = EncoderFormat
        gzip = EncoderGzip
        json = EncoderJson
        newline = EncoderNewline
        prefix = EncoderPrefix
        suffix = EncoderSuffix

        # todo consider adding msgpack encoder

    def set_basic_params(
            self, no_requests=None, template=None, memory_report=None, prefix=None, prefix_date=None,
            apply_strftime=None, response_ms=None, ip_x_forwarded=None):
        """

        :param bool no_requests: Disable requests logging - only uWSGI internal messages
            and errors will be logged.

        :param str|unicode template: Set advanced format for request logging.
            This template string can use variables from ``Logging.Vars``.

        :param str|unicode prefix: Prefix log items with a string.

        :param str|unicode prefix_date: Prefix log items with date.

        :param int memory_report: Enable memory report.
                * **1** - basic (default);
                * **2** - uss/pss (Linux only)

        :param bool apply_strftime: Apply strftime to logformat output.

        :param bool response_ms: Report response time in microseconds instead of milliseconds.

        :param bool ip_x_forwarded: Use the IP from X-Forwarded-For header instead of REMOTE_ADDR.
            Used when uWSGI is run behind multiple proxies.

        """
        self._set('disable-logging', no_requests, cast=bool)
        self._set('log-format', template)
        self._set('memory-report', memory_report)
        self._set('log-prefix', prefix)
        self._set('log-date', prefix_date, cast=bool)
        self._set('logformat-strftime', apply_strftime, cast=bool)
        self._set('log-micros', response_ms, cast=bool)
        self._set('log-x-forwarded-for', ip_x_forwarded, cast=bool)

        return self._section

    def set_file_params(
            self, reopen_on_reload=None, trucate_on_statup=None, max_size=None, rotation_fname=None,
            touch_reopen=None, touch_rotate=None, owner=None, mode=None):
        """

        :param bool reopen_on_reload: Reopen log after reload.

        :param bool trucate_on_statup: Truncate log on startup.

        :param int max_size: Set maximum logfile size in bytes.

        :param str|unicode rotation_fname: Set log file name after rotation.

        :param str|unicode|list touch_reopen: Trigger log reopen if the specified file
            is modified/touched.

        :param str|unicode|list touch_rotate: Trigger log rotation if the specified file
            is modified/touched.

        :param str|unicode owner: Set owner chown() for logs.
        .
        :param str|unicode mode: Set mode chmod() for logs.

        """
        self._set('log-reopen', reopen_on_reload, cast=bool)
        self._set('log-truncate', trucate_on_statup, cast=bool)
        self._set('log-maxsize', max_size)
        self._set('log-backupname', rotation_fname)

        self._set('touch-logreopen', touch_reopen, multi=True)
        self._set('touch-logrotate', touch_rotate, multi=True)

        self._set('logfile-chown', owner)
        self._set('logfile-chmod', mode)

        return self._section

    def set_filters(
            self, include=None, exclude=None, slower=None, bigger=None, status_4xx=None, status_5xx=None,
            no_body=None, sendfile=None, io_errors=None):
        """

        :param str|unicode|list include: Show only log lines matching the specified regexp.

        :param str|unicode|list exclude: Do not show log lines matching the specified regexp.

        :param int slower: Log requests slower than the specified number of milliseconds.

        :param int bigger: Log requestes bigger than the specified size in bytes.

        :param status_4xx: Log requests with a 4xx response.

        :param status_5xx: Log requests with a 5xx response.

        :param bool no_body: Log responses without body.

        :param bool sendfile: Log sendfile requests.

        :param bool io_errors: Log requests with io errors.

        """
        self._set('log-slow', slower)
        self._set('log-big', bigger)
        self._set('log-4xx', status_4xx, cast=bool)
        self._set('log-5xx', status_5xx, cast=bool)
        self._set('log-zero', no_body, cast=bool)
        self._set('log-sendfile', sendfile, cast=bool)
        self._set('log-ioerror', io_errors, cast=bool)

        for line in listify(include):
            self._set('log-filter', line, multi=True)

        for line in listify(exclude):
            self._set('log-drain', line, multi=True)

        return self._section

    def set_master_logging_params(
            self, enable=None, dedicate_thread=None, buffer=None,
            sock_stream=None, sock_stream_requests_only=None):
        """Sets logging params for delegating logging to master process.

        :param bool enable: Delegate logging to master process.
            Delegate the write of the logs to the master process
            (this will put all of the logging I/O to a single process).
            Useful for system with advanced I/O schedulers/elevators.

        :param bool dedicate_thread: Delegate log writing to a thread.

            As error situations could cause the master to block while writing
            a log line to a remote server, it may be a good idea to use this option and delegate
            writes to a secondary thread.

        :param int buffer: Set the buffer size for the master logger in bytes.
            Bigger log messages will be truncated.

        :param bool|tuple sock_stream: Create the master logpipe as SOCK_STREAM.

        :param bool|tuple sock_stream_requests_only: Create the master requests logpipe as SOCK_STREAM.

        """
        self._set('log-master', enable, cast=bool)
        self._set('threaded-logger', dedicate_thread, cast=bool)
        self._set('log-master-bufsize', buffer)

        self._set('log-master-stream', sock_stream, cast=bool)

        if sock_stream_requests_only:
            self._set('log-master-req-stream', sock_stream_requests_only, cast=bool)

        return self._section

    def print_loggers(self):
        """Print out available (built) loggers."""

        self._set('loggers-list', True, cast=bool)

        return self._section

    def add_logger(self, logger, requests_only=False, for_single_worker=False):
        """Set/add a common logger or a request requests only.

        :param str|unicode|list|Logger|list[Logger] logger:

        :param bool requests_only: Logger used only for requests information messages.

        :param bool for_single_worker: Logger to be used in single-worker setup.


        """
        if for_single_worker:
            command = 'worker-logger-req' if requests_only else 'worker-logger'
        else:
            command = 'req-logger' if requests_only else 'logger'

        for logger in listify(logger):
            self._set(command, logger, multi=True)

        return self._section

    def add_logger_route(self, logger, matcher, requests_only=False):
        """Log to the specified named logger if regexp applied on log item matches.

        :param str|unicode|list|Logger|list[Logger] logger: Logger to associate route with.

        :param str|unicode matcher: Regular expression to apply to log item.

        :param bool requests_only: Matching should be used only for requests information messages.

        """
        command = 'log-req-route' if requests_only else 'log-route'

        for logger in listify(logger):
            self._set(command, '%s %s' % (logger, matcher), multi=True)

        return self._section

    def add_logger_encoder(self, encoder, logger=None, requests_only=False, for_single_worker=False):
        """Add an item in the log encoder or request encoder chain.

        * http://uwsgi-docs.readthedocs.io/en/latest/LogEncoders.html

            .. note:: Encoders automatically enable master log handling (see ``.set_master_logging_params()``).

            .. note:: For best performance consider allocating a thread
                for log sending with ``dedicate_thread``.

        :param str|unicode|list|Encoder encoder: Encoder (or a list) to add into processing.

        :param str|unicode|Logger logger: Logger apply associate encoders to.

        :param bool requests_only: Encoder to be used only for requests information messages.

        :param bool for_single_worker: Encoder to be used in single-worker setup.

        """
        if for_single_worker:
            command = 'worker-log-req-encoder' if requests_only else 'worker-log-encoder'
        else:
            command = 'log-req-encoder' if requests_only else 'log-encoder'

        for encoder in listify(encoder):

            value = '%s' % encoder

            if logger:
                if isinstance(logger, Logger):
                    logger = logger.alias

                value += ':%s' % logger

            self._set(command, value, multi=True)

        return self._section

    class vars(object):
        """Variables available for custom log formatting."""

        # The following are taken blindly from the internal wsgi_request structure of the current request.

        REQ_URL = '%(uri)'
        """REQUEST_URI from ``wsgi_request`` of the current request."""

        REQ_METHOD = '%(method)'
        """REQUEST_METHOD from ``wsgi_request`` of the current request."""

        REQ_REMOTE_USER = '%(user)'
        """REMOTE_USER from ``wsgi_request`` of the current request."""

        REQ_REMOTE_ADDR = '%(addr)'
        """REMOTE_ADDR from ``wsgi_request`` of the current request."""

        REQ_HTTP_HOST = '%(host)'
        """HTTP_HOST from ``wsgi_request`` of the current request."""

        REQ_SERVER_PROTOCOL = '%(proto)'
        """SERVER_PROTOCOL from ``wsgi_request`` of the current request."""

        REQ_USER_AGENT = '%(uagent)'
        """HTTP_USER_AGENT from ``wsgi_request`` of the current request."""

        REQ_REFERER = '%(referer)'
        """HTTP_REFERER from ``wsgi_request`` of the current request."""

        # The following are simple functions called for generating the logvar value.

        RESP_STATUS = '%(status)'
        """HTTP response status code."""

        RESP_TIME_US = '%(micros)'
        """Response time in microseconds."""

        RESP_TIME_MS = '%(msecs)'
        """Response time in milliseconds."""

        REQ_START_TS = '%(time)'
        """Timestamp of the start of the request."""

        RESP_START_CTIME = '%(ctime)'
        """Ctime of the start of the request."""

        TIME_UNIX = '%(epoch)'
        """The current time in Unix format."""

        RESP_SIZE = '%(size)'
        """Response body size + response headers size."""

        REQ_TIME_HUMAN = '%(ltime)'
        """Human-formatted (Apache style) request time."""

        REQ_TIME_FORMATTED = '%(ftime)'
        """Request time formatted with ``date_format``."""

        RESP_SIZE_HEADER = '%(hsize)'
        """Response headers size."""

        RESP_SIZE_BODY = '%(rsize)'
        """Response body size."""

        REQ_SIZE_BODY = '%(cl)'
        """Request content body size."""

        WORKER_PID = '%(pid)'
        """pid of the worker handling the request."""

        WORKER_ID = '%(wid)'
        """id of the worker handling the request."""

        ASYNC_SWITCHES = '%(switches)'
        """Number of async switches."""

        REQ_COUNT_VARS_CGI = '%(vars)'
        """Number of CGI vars in the request."""

        RESP_COUNT_HEADERS = '%(headers)'
        """Number of generated response headers."""

        CORE = '%(core)'
        """The core running the request."""

        MEM_VSZ = '%(vsz)'
        """Address space/virtual memory usage (in bytes)."""

        MEM_RSS = '%(rss)'
        """RSS memory usage (in bytes)."""

        MEM_VSZ_MB = '%(vszM)'
        """Address space/virtual memory usage (in megabytes)."""

        MEM_RSS_MV = '%(rssM)'
        """RSS memory usage (in megabytes)."""

        SIZE_PACKET_UWSGI = '%(pktsize)'
        """Size of the internal request uwsgi packet."""

        MOD1 = '%(modifier1)'
        """modifier1 of the request."""

        MOD2 = '%(modifier2)'
        """modifier2 of the request."""

        REQ_COUNT_ERR_READ = '%(rerr)'
        """Number of read errors for the request.
    
        .. note:: since 1.9.21
    
        """

        REQ_COUNT_ERR_WRITE = '%(werr)'
        """Number of write errors for the request.
    
        .. note:: since 1.9.21
    
        """

        REQ_COUNT_ERR = '%(ioerr)'
        """Number of write and read errors for the request.
    
        .. note:: since 1.9.21
    
        """

        REQ_START_UNIX_US = '%(tmsecs)'
        """Timestamp of the start of the request in milliseconds since the epoch.
    
        .. note:: since 1.9.21
    
        """

        REQ_START_UNIX_MS = '%(tmicros)'
        """Timestamp of the start of the request in microseconds since the epoch.
    
        .. note:: since 1.9.21
    
        """

        # todo %(metric.XXX) - access the XXX metric value (see The Metrics subsystem)
        # todo %(var.XXX)





def listify(src):
    """Make a list with source object if not already a list.

    :param src:
    :rtype: list

    """
    if not isinstance(src, list):
        src = [src]

    return src


def filter_locals(locals_dict, drop=None):
    """Filters a dictionary produced by locals().

    :param dict locals_dict:

    :param list drop: Keys to drop from dict.

    :rtype: dict
    """
    drop = drop or []
    drop.extend([
        'self',
        '__class__',  # py3
    ])
    locals_dict = {k: v for k, v in locals_dict.items() if k not in drop}
    return locals_dict


def make_key_val_string(
        locals_dict, keys=None, aliases=None, bool_keys=None, list_keys=None,
        items_separator=','):
    """Flattens the given dictionary into key-value string.

    :param dict locals_dict: Dictionary produced by locals().

    :param list keys: Relevant keys from dictionary.
        If not defined - all keys are relevant.
        If defined keys will flattened into string using given order.

    :param dict aliases: Mapping key names from locals_dict into names
        they should be replaced with.

    :param list bool_keys: Keys to consider their values bool.

    :param list list_keys: Keys expecting lists.

    :param str|unicode items_separator: String to use as items (chunks) separator.

    :rtype: str|unicode
    """
    value_chunks = []

    keys = keys or sorted(filter_locals(locals_dict).keys())
    aliases = aliases or {}
    bool_keys = bool_keys or []
    list_keys = list_keys or []

    for key in keys:
        val = locals_dict[key]

        if val is not None:

            if key in bool_keys:
                val = 1

            elif key in list_keys:

                val = ';'.join(listify(val))

            value_chunks.append('%s=%s' % (aliases.get(key, key), val))

    return items_separator.join(value_chunks).strip()



from ..base import OptionsGroup
from ..utils import listify


class Statics(OptionsGroup):
    """Statics.

    Unfortunately you cannot live without serving static files via some protocol (HTTP, SPDY or something else).
    Fortunately uWSGI has a wide series of options and micro-optimizations for serving static files.

    .. note:: This subsystem automatically honours the ``If-Modified-Since`` HTTP request header.

    * http://uwsgi.readthedocs.io/en/latest/StaticFiles.html

    """

    DIR_DOCUMENT_ROOT = 'docroot'
    """Used to check for static files in the requested DOCUMENT_ROOT. Pass into ``static_dir``."""

    class expiration_criteria(object):
        """Expiration criteria (subjects) to use with ``.add_expiration_rule()``."""

        FILENAME = 'filename'
        """Allows setting the Expires header for the specified file name pattern."""

        MIME_TYPE = 'type'
        """Allows setting the Expires header for the specified MIME type."""

        PATH_INFO = 'path-info'
        """Allows setting the Expires header for the specified ``PATH_INFO`` pattern."""

        REQUEST_URI = 'uri'
        """Allows setting the Expires header for the specified ``REQUEST_URI`` pattern."""

    class transfer_modes(object):
        """File transfer (serving) modes.

        With this, uWSGI will only generate response headers and
        the web server will be delegated to transferring the physical file.

        * http://uwsgi.readthedocs.io/en/latest/StaticFiles.html#transfer-modes

        """

        SENDFILE = 'x-sendfile'
        """Use ``X-Sendfile`` mode. Apache."""

        ACCEL_REDIRECT = 'x-accel-redirect'
        """Use ``X-Accel-Redirect`` mode. Nginx."""

    def set_basic_params(
            self, static_dir=None, index_file=None, mime_file=None, skip_ext=None, transfer_mode=None):
        """

        :param str|unicode|list[str|unicode] static_dir: Check for static files in the specified directory.

            .. note:: Use ``DIR_DOCUMENT_ROOT`` constant to serve files under ``DOCUMENT_ROOT``.

        :param str|unicode|list[str|unicode] index_file: Search for specified file if a directory is requested.

            Example: ``index.html``

        :param str|unicode|list[str|unicode] mime_file: Set mime types file path to extend uWSGI builtin list.

            Default: ``/etc/mime.types`` or ``/etc/apache2/mime.types``.

        :param str|unicode|list[str|unicode] skip_ext: Skip specified extension from static file checks.

            Example: add ``.php`` to not serve it as static.

        :param str|unicode transfer_mode: Set static file serving (transfer) mode.

            See ``.transfer_modes``.

            .. note:: Another option is to specify ``count_offload`` in ``.workers.set_thread_params()``.

        """

        if static_dir == self.DIR_DOCUMENT_ROOT:
            self._set('check-static-docroot', True, cast=bool)

        else:
            self._set('static-check', static_dir, multi=True)

        self._set('static-index', index_file, multi=True)
        self._set('mimefile', mime_file, multi=True)
        self._set('static-skip-ext', skip_ext, multi=True)
        self._set('fileserve-mode', skip_ext, multi=True)

        return self._section

    def register_static_map(self, mountpoint, target, retain_resource_path=False, safe_target=False):
        """Allows mapping mountpoint to a static directory (or file).

        * http://uwsgi.readthedocs.io/en/latest/StaticFiles.html#mode-3-using-static-file-mount-points

        :param str|unicode mountpoint:

        :param str|unicode target:

        :param bool retain_resource_path: Append the requested resource to the docroot.

            Example: if ``/images`` maps to ``/var/www/img`` requested ``/images/logo.png`` will be served from:

            * ``True``: ``/var/www/img/images/logo.png``

            * ``False``: ``/var/www/img/logo.png``

        :param bool safe_target: Skip security checks if the file is under the specified path.

            Whether to consider resolved (real) target a safe one to serve from.

            * http://uwsgi.readthedocs.io/en/latest/StaticFiles.html#security

        """
        command = 'static-map'

        if retain_resource_path:

            command += '2'

        self._set(command, '%s=%s' % (mountpoint, target), multi=True)

        if safe_target:
            self._set('static-safe', target, multi=True)

        return self._section

    def add_expiration_rule(self, criterion, value, timeout, use_mod_time=False):
        """Adds statics expiration rule based on a criterion.

        :param str|unicode criterion: Criterion (subject) to base expiration on.

            See ``.expiration_criteria``.

        :param str|unicode|list[str|unicode] value: Value to test criteria upon.

            .. note:: Usually a regular expression.

        :param int timeout: Number of seconds to expire after.

        :param bool use_mod_time: Base on file modification time instead of the current time.

        """
        command = 'static-expires'
        separator = ' '

        if criterion != self.expiration_criteria.FILENAME:

            command += '-%s' % criterion

        if criterion == self.expiration_criteria.MIME_TYPE:

            separator = '='

        if use_mod_time:

            command += '-mtime'

        for value in listify(value):
            self._set(command, '%s%s%s' % (value, separator, timeout), multi=True)

        return self._section

    # todo consider adding:
    # static-gzip*

    def set_paths_caching_params(self, timeout=None, cache_name=None):
        """Use the uWSGI caching subsystem to store mappings from URI to filesystem paths.

        * http://uwsgi.readthedocs.io/en/latest/StaticFiles.html#caching-paths-mappings-resolutions

        :param int timeout: Amount of seconds to put resolved paths in the uWSGI cache.

        :param str|unicode cache_name: Cache name to use for static paths.

        """
        self._set('static-cache-paths', timeout)
        self._set('static-cache-paths-name', cache_name)

        return self._section



from ..base import OptionsGroup
from ..utils import make_key_val_string


class Caching(OptionsGroup):
    """Caching.

    uWSGI includes a very fast, all-in-memory, zero-IPC, SMP-safe, constantly-optimizing,
    highly-tunable, key-value store simply called "the caching framework".

    A single uWSGI instance can create an unlimited number of "caches" each one
    with different setup and purpose.

    * http://uwsgi-docs.readthedocs.io/en/latest/Caching.html
    * http://uwsgi-docs.readthedocs.io/en/latest/tutorials/CachingCookbook.html

    """

    def set_basic_params(self, no_expire=None, expire_scan_interval=None, report_freed=None):
        """
        :param bool no_expire: Disable auto sweep of expired items.
            Since uWSGI 1.2, cache item expiration is managed by a thread in the master process,
            to reduce the risk of deadlock. This thread can be disabled
            (making item expiry a no-op) with the this option.

        :param int expire_scan_interval: Set the frequency (in seconds) of cache sweeper scans. Default: 3.

        :param bool report_freed: Constantly report the cache item freed by the sweeper.

            .. warning:: Use only for debug.

        """
        self._set('cache-no-expire', no_expire, cast=bool)
        self._set('cache-report-freed-items', report_freed, cast=bool)
        self._set('cache-expire-freq', expire_scan_interval)

        return self._section

    def add_item(self, key, value, cache_name=None):
        """Add an item into the given cache.

        This is a commodity option (mainly useful for testing) allowing you
        to store an item in a uWSGI cache during startup.

        :param str|unicode key:

        :param value:

        :param str|unicode cache_name: If not set, default will be used.

        """
        cache_name = cache_name or ''
        value = '%s %s=%s' % (cache_name, key, value)

        self._set('add-cache-item', value.strip(), multi=True)

        return self._section

    def add_file(self, filepath, gzip=False, cache_name=None):
        """Load a static file in the cache.

        .. note:: Items are stored with the filepath as is (relative or absolute) as the key.

        :param str|unicode filepath:

        :param bool gzip: Use gzip compression.

        :param str|unicode cache_name: If not set, default will be used.

        """
        command = 'load-file-in-cache'

        if gzip:
            command += '-gzip'

        cache_name = cache_name or ''
        value = '%s %s' % (cache_name, filepath)

        self._set(command, value.strip(), multi=True)

        return self._section

    def add_cache(
            self, name, max_items, expires=None, store=None, store_sync_interval=None, store_delete=None,
            hash_algo=None, hash_size=None, key_size=None, udp_clients=None, udp_servers=None,
            block_size=None, block_count=None, sync_from=None, mode_bitmap=None, use_lastmod=None,
            full_silent=None, full_purge_lru=None):
        """Creates cache. Default mode: single block.

        .. note:: This uses new generation ``cache2`` option available since uWSGI 1.9.

        .. note:: When at least one cache is configured without ``full_purge_lru``
            and the master is enabled a thread named "the cache sweeper" is started.
            Its main purpose is deleting expired keys from the cache.
            If you want auto-expiring you need to enable the master.

        :param str|unicode name: Set the name of the cache. Must be unique in an instance.

        :param int max_items: Set the maximum number of cache items.

            .. note:: Effective number of items is **max_items - 1** -
                the first item of the cache is always internally used as "NULL/None/undef".

        :param int expires: The number of seconds after the object is no more valid
            (and will be removed by the cache sweeper when ``full_purge_lru`` is not set.

        :param str|unicode store: Set the filename for the persistent storage.
            If it doesn't exist, the system assumes an empty cache and the file will be created.

        :param int store_sync_interval: Set the number of seconds after which msync() is called
            to flush memory cache on disk when in persistent mode.
            By default it is disabled leaving the decision-making to the kernel.

        :param bool store_delete: uWSGI, by default, will not start if a cache file exists
            and the store file does not match the configured items/blocksize.
            Setting this option will make uWSGI delete the existing file upon mismatch
            and create a new one.

        :param str|unicode hash_algo: Set the hash algorithm used in the hash table. Current options are:

            * djb33x (default)
            * murmur2

        :param int hash_size: This is the size of the hash table in bytes.
            Generally 65536 (the default) is a good value.

            .. note:: Change it only if you know what you are doing
                or if you have a lot of collisions in your cache.

        :param int key_size: Set the maximum size of a key, in bytes. Default: 2048.

        :param str|unicode|list udp_clients: List of UDP servers which will receive UDP cache updates.

        :param str|unicode |list udp_servers: List of UDP addresses on which to bind the cache
            to wait for UDP updates.

        :param int block_size: Set the size (in bytes) of a single block.

            .. note:: It's a good idea to use a multiple of 4096 (common memory page size).

        :param int block_count: Set the number of blocks in the cache. Useful only in bitmap mode,
            otherwise the number of blocks is equal to the maximum number of items.

        :param str|unicode|list sync_from: List of uWSGI addresses which the cache subsystem will connect to
            for getting a full dump of the cache. It can be used for initial cache synchronization.
            The first node sending a valid dump will stop the procedure.
            
        :param bool mode_bitmap: Enable (more versatile but relatively slower) bitmap mode.

            http://uwsgi-docs.readthedocs.io/en/latest/Caching.html#single-block-faster-vs-bitmaps-slower

            .. warning:: Considered production ready only from uWSGI 2.0.2.

        :param bool use_lastmod: Enabling will update last_modified_at timestamp of each cache
            on every cache item modification. Enable it if you want to track this value
            or if other features depend on it. This value will then be accessible via the stats socket.

        :param bool full_silent: By default uWSGI will print warning message on every cache set operation
            if the cache is full. To disable this warning set this option.

            .. note:: Available since 2.0.4.

        :param bool full_purge_lru: Allows the caching framework to evict Least Recently Used (LRU)
            item when you try to add new item to cache storage that is full.

            .. note:: ``expires`` argument will be ignored.

        """
        value = make_key_val_string(
            locals(),
            keys=[
                'name', 'max_items', 'expires', 'store', 'store_sync_interval', 'store_delete',
                'hash_algo', 'hash_size', 'key_size', 'udp_clients', 'udp_servers',
                'block_size', 'block_count', 'sync_from', 'mode_bitmap', 'use_lastmod',
                'full_silent', 'full_purge_lru',
            ],
            aliases={
                'max_items': 'maxitems',
                'store_sync_interval': 'storesync',
                'hash_algo': 'hash',
                'udp_clients': 'nodes',
                'block_size': 'blocksize',
                'block_count': 'blocks',
                'sync_from': 'sync',
                'mode_bitmap': 'bitmap',
                'use_lastmod': 'lastmod',
                'full_silent': 'ignore_full',
                'full_purge_lru': 'purge_lru',
            },
            bool_keys=['store_delete', 'mode_bitmap', 'use_lastmod', 'full_silent', 'full_purge_lru'],
            list_keys=['udp_clients', 'udp_servers', 'sync_from'],
        )

        self._set('cache2', value, multi=True)

        return self._section



from uwsgiconf import Section


def test_empire_basics(assert_lines):

    assert_lines([
        'emperor = /myvassals',
    ], Section().empire.set_emperor_params(vassals_home='/myvassals'))

    assert_lines([
        'imperial-monitor-list = true',
    ], Section().empire.print_monitors())

    assert_lines([
        'emperor-command-socket = /mysock',
    ], Section().empire.set_emperor_command_params('/mysock'))

    assert_lines([
        'emperor-wrapper = myvas',
    ], Section().empire.set_vassals_wrapper_params('myvas'))

    assert_lines([
        'emperor-throttle = 100',
    ], Section().empire.set_throttle_params(100))

    assert_lines([
        'emperor-required-heartbeat = 200',
    ], Section().empire.set_tolerance_params(200))

    assert_lines([
        'emperor-tyrant = true',
    ], Section().empire.set_mode_tyrant_params(enable=True))

    assert_lines([
        'emperor-broodlord = 40',
    ], Section().empire.set_mode_broodlord_params(40))



from ..base import OptionsGroup
from .. utils import make_key_val_string
from ..exceptions import ConfigurationError


class MasterProcess(OptionsGroup):
    """Master process is a separate process offering mentoring capabilities
    for other processes. Only one master process per uWSGI instance.

    uWSGI's built-in prefork+threading multi-worker management mode,
    activated by flicking the master switch on.

    .. note:: For all practical serving deployments it's not really a good idea not to use master mode.

    """

    def set_basic_params(
            self, enable=None, name=None, no_orphans=None, as_root=None,
            subproc_check_interval=None, fifo_file=None):
        """

        :param bool enable: Enable uWSGI master process.

        :param str|unicode master_enabled: Set master process name to given value.

        :param bool no_orphans: Automatically kill workers if master dies (can be dangerous for availability).

        :param bool as_root: Leave master process running as root.

        :param int subproc_check_interval: Set the interval (in seconds) of master checks. Default: 1
            The master process makes a scan of subprocesses, etc. every N seconds.

            .. warning:: You can increase this time if you need to, but it's DISCOURAGED.

        :param str|unicode fifo_file: Enables the master FIFO.

            Instead of signals, you can tell the master to create a UNIX named pipe (FIFO)
            that you may use to issue commands to the master.

            Up to 10 different FIFO files supported. By default the first specified is bound (mapped as '0').

            * http://uwsgi.readthedocs.io/en/latest/MasterFIFO.html#the-master-fifo

            .. note:: Since 1.9.17

        """
        self._set('master', enable, cast=bool)
        self._set('procname-master', name)
        self._set('no-orphans', no_orphans)
        self._set('master-as-root', as_root)
        self._set('check-interval', subproc_check_interval)
        self._set('master-fifo', fifo_file, multi=True)

        return self._section

    def set_exit_events(self, no_workers=None, idle=None, reload=None):
        """Do exit on certain events

        :param bool no_workers: Shutdown uWSGI when no workers are running.

        :param bool idle: Shutdown uWSGI when idle.

        :param bool reload: Force exit even if a reload is requested.

        """
        self._set('die-on-no-workers', no_workers, cast=bool)
        self._set('exit-on-reload', reload, cast=bool)
        self.set_idle_params(exit=idle)

    def set_idle_params(self, timeout=None, exit=None):
        """Activate idle mode - put uWSGI in cheap mode after inactivity timeout.

        :param int timeout: Inactivity timeout in seconds.

        :param bool exit: Shutdown uWSGI when idle.

        """
        self._set('idle', timeout)
        self._set('die-on-idle', exit, cast=bool)

        return self._section

    def set_reload_params(self, mercy=None, exit=None):
        """Set reload related params.

        :param int mercy: Set the maximum time (in seconds) we wait
            for workers and other processes to die during reload/shutdown.

        :param bool exit: Force exit even if a reload is requested.

        """
        self._set('reload-mercy', mercy)
        self.set_exit_events(reload=exit)

        return self._section

    def add_cron_task(
            self, command, weekday=None, month=None, day=None, hour=None, minute=None,
            legion=None, unique=None, harakiri=None):
        """Adds a cron task running the given command on the given schedule.
        http://uwsgi.readthedocs.io/en/latest/Cron.html

        HINTS:
            * Use negative values to say `every`:
                hour=-3  stands for `every 3 hours`

            * Use - (minus) to make interval:
                minute='13-18'  stands for `from minute 13 to 18`

        .. note:: We use cron2 option available since 1.9.11.

        :param str|unicode command: Command to execute on schedule (with or without path).

        :param int|str|unicode weekday: Day of a the week number. Defaults to `each`.
            0 - Sunday  1 - Monday  2 - Tuesday  3 - Wednesday
            4 - Thursday  5 - Friday  6 - Saturday

        :param int|str|unicode month: Month number 1-12. Defaults to `each`.

        :param int|str|unicode day: Day of the month number 1-31. Defaults to `each`.

        :param int|str|unicode hour: Hour 0-23. Defaults to `each`.

        :param int|str|unicode minute: Minute 0-59. Defaults to `each`.

        :param str|unicode legion: Set legion (cluster) name to use this cron command against.
            Such commands are only executed by legion lord node.

        :param bool unique: Marks command as unique. Default to not unique.
            Some commands can take a long time to finish or just hang doing their thing.
            Sometimes this is okay, but there are also cases when running multiple instances
            of the same command can be dangerous.

        :param int harakiri: Enforce a time limit (in seconds) on executed commands.
            If a command is taking longer it will be killed.

        """
        rule = make_key_val_string(
            locals(),
            keys=['weekday', 'month', 'day', 'hour', 'minute', 'harakiri', 'legion', 'unique'],
            aliases={'weekday': 'week'},
            bool_keys=['unique'],
        )

        self._set('cron2', ('%s %s' % (rule, command)).strip(), multi=True)

        return self._section

    def attach_process_classic(self, command_or_pid_path, background, control=False, for_legion=False):
        """Attaches a command/daemon to the master process optionally managed by a pidfile.

        This will allow the uWSGI master to control/monitor/respawn this process.

        .. note:: This uses old classic uWSGI means of process attaching
            To have more control use ``.attach_process()`` method (requires  uWSGI 2.0+)

        http://uwsgi-docs.readthedocs.io/en/latest/AttachingDaemons.html

        :param str|unicode command_or_pid_path:

        :param bool background: Must indicate whether process is in background.

        :param bool control: Consider this process a control: when the daemon dies, the master exits.

            .. note:: pidfile managed processed not supported.

        :param bool for_legion: Legion daemons will be executed only on the legion lord node,
            so there will always be a single daemon instance running in each legion.
            Once the lord dies a daemon will be spawned on another node.

            .. note:: uWSGI 1.9.9+ required.

        """
        prefix = 'legion-' if for_legion else ''

        if '.pid' in command_or_pid_path:

            if background:
                # Attach a command/daemon to the master process managed by a pidfile (the command must daemonize)
                self._set(prefix + 'smart-attach-daemon', command_or_pid_path, multi=True)

            else:
                # Attach a command/daemon to the master process managed by a pidfile (the command must NOT daemonize)
                self._set(prefix + 'smart-attach-daemon2', command_or_pid_path, multi=True)

        else:
            if background:
                raise ConfigurationError('Background flag is only supported for pid-governed commands')

            if control:
                # todo needs check
                self._set('attach-control-daemon', command_or_pid_path, multi=True)

            else:
                # Attach a command/daemon to the master process (the command has to remain in foreground)
                self._set(prefix + 'attach-daemon', command_or_pid_path, multi=True)

        return self._section

    def attach_process(
            self, command, for_legion=False, broken_counter=None, pidfile=None, control=None, daemonize=None,
            touch_reload=None, signal_stop=None, signal_reload=None, honour_stdin=None,
            uid=None, gid=None, new_pid_ns=None, change_dir=None):
        """Attaches a command/daemon to the master process.

        This will allow the uWSGI master to control/monitor/respawn this process.

        http://uwsgi-docs.readthedocs.io/en/latest/AttachingDaemons.html

        :param str|unicode command: The command line to execute.

        :param bool for_legion: Legion daemons will be executed only on the legion lord node,
            so there will always be a single daemon instance running in each legion.
            Once the lord dies a daemon will be spawned on another node.

        :param int broken_counter: Maximum attempts before considering a daemon "broken".

        :param str|unicode pidfile: The pidfile path to check (enable smart mode).

        :param bool control: If True, the daemon becomes a `control` one:
            if it dies the whole uWSGI instance dies.

        :param bool daemonize: Daemonize the process (enable smart2 mode).

        :param list|str|unicode touch_reload: List of files to check:
            whenever they are 'touched', the daemon is restarted

        :param int signal_stop: The signal number to send to the daemon when uWSGI is stopped.

        :param int signal_reload: The signal number to send to the daemon when uWSGI is reloaded.

        :param bool honour_stdin: The signal number to send to the daemon when uWSGI is reloaded.

        :param str|unicode|int uid: Drop privileges to the specified uid.

            .. note:: Requires master running as root.

        :param str|unicode|int gid: Drop privileges to the specified gid.

            .. note:: Requires master running as root.

        :param bool new_pid_ns: Spawn the process in a new pid namespace.

            .. note:: Requires master running as root.

            .. note:: Linux only.

        :param str|unicode change_dir: Use chdir() to the specified directory
            before running the command.

        """
        rule = make_key_val_string(
            locals(),
            keys=[
                'command', 'broken_counter', 'pidfile', 'control', 'daemonize', 'touch_reload',
                'signal_stop', 'signal_reload', 'honour_stdin',
                'uid', 'gid', 'new_pid_ns', 'change_dir',
            ],
            aliases={
                'command': 'cmd',
                'broken_counter': 'freq',
                'touch_reload': 'touch',
                'signal_stop': 'stopsignal',
                'signal_reload': 'reloadsignal',
                'honour_stdin': 'stdin',
                'new_pid_ns': 'ns_pid',
                'change_dir': 'chdir',
            },
            bool_keys=['control', 'daemonize', 'honour_stdin'],
            list_keys=['touch_reload'],
        )

        prefix = 'legion-' if for_legion else ''

        self._set(prefix + 'attach-daemon2', rule, multi=True)

        return self._section



from uwsgiconf.presets.nice import Section, PythonSection


def test_nice_section(assert_lines):

    assert_lines([
        'workers = %k',

    ], Section())

    assert_lines([
        'workers = 13',
        'touch-reload', 'test_nice.py',

    ], Section(workers=13, touch_reload=__file__))


def test_nice_python(assert_lines):

    assert_lines([
        'plugin = python',
        'wsgi = somepackage.module',

    ], PythonSection(wsgi_module='somepackage.module'))



from uwsgiconf import Section


def test_alarms_basics(assert_lines):

    assert_lines([
        'alarm-msg-size = 10000',
    ], Section().alarms.set_basic_params(msg_size=10000))

    assert_lines([
        'alarm-list = true',
    ], Section().alarms.print_alarms())


def test_alarms_registration(assert_lines):

    alarms = Section().alarms
    assert_lines([
        'alarm = my log:',
    ], alarms.register_alarm(alarms.alarm_types.log('my')))

    alarms = Section().alarms
    assert_lines([
        'alarm = my cmd:some',
    ], alarms.register_alarm(alarms.alarm_types.command('my', 'some')))

    alarms = Section().alarms
    assert_lines([
        'alarm = mysig signal:17',
    ], alarms.register_alarm(alarms.alarm_types.signal('mysig', 17)))

    alarms = Section().alarms
    assert_lines([
        'alarm = tomule mule:2',
    ], alarms.register_alarm(alarms.alarm_types.mule('tomule', 2)))

    alarms = Section().alarms
    assert_lines([
        'plugin = alarm_curl',
        'alarm = test2 curl:http://192.168.173.6:9191/argh;auth_pass=foobar;auth_user=topogigio',
    ], alarms.register_alarm(
        alarms.alarm_types.curl(
            'test2', 'http://192.168.173.6:9191/argh',
            auth_user='topogigio', auth_pass='foobar')))

    alarms = Section().alarms
    assert_lines([
        'plugin = alarm_xmpp',
        'alarm = jab xmpp:idle@some.com;12345;one@some.com,two@some.com',
    ], alarms.register_alarm(
        alarms.alarm_types.xmpp('jab', 'idle@some.com', '12345', ['one@some.com', 'two@some.com'])))


def test_alarms_on_log(assert_lines):

    alarms = Section().alarms

    alarm1 = alarms.alarm_types.command('mycom', 'some')
    alarm2 = alarms.alarm_types.signal('mysig', 27)

    assert_lines([
        'alarm = mycom cmd:some',
        'alarm = mysig signal:27',
        'alarm-log = mycom,mysig some text',
    ], alarms.alarm_on_log([alarm1, alarm2], 'some text'))

    assert_lines([
        'not-alarm-log = mycom other',
    ], alarms.alarm_on_log(alarm1, 'other', skip=True))


def test_alarms_on_fd(assert_lines):

    alarms = Section().alarms
    alarm1 = alarms.alarm_types.signal('mysig', 27)
    alarm2 = alarms.alarm_types.signal('some', 17)

    assert_lines([
        'alarm = mysig signal:27',
        'alarm = some signal:17',
        'alarm-fd = mysig $(CGROUP_OOM_FD):8 damn it!',
        'alarm-fd = some $(CGROUP_OOM_FD):8 damn it!',
    ], alarms.alarm_on_fd_ready([alarm1, alarm2], '$(CGROUP_OOM_FD)', 'damn it!', byte_count=8))


def test_alarms_on_backlog(assert_lines):

    alarms = Section().alarms
    alarm1 = alarms.alarm_types.signal('mysig', 27)
    alarm2 = alarms.alarm_types.signal('some', 17)

    assert_lines([
        'alarm = mysig signal:27',
        'alarm = some signal:17',
        'alarm-backlog = mysig',
        'alarm-backlog = some',
    ], alarms.alarm_on_backlog_full([alarm1, alarm2]))


def test_alarms_on_segfault(assert_lines):

    alarms = Section().alarms
    alarm1 = alarms.alarm_types.signal('mysig', 27)
    alarm2 = alarms.alarm_types.signal('some', 17)

    assert_lines([
        'alarm = mysig signal:27',
        'alarm = some signal:17',
        'alarm-segfault = mysig',
        'alarm-segfault = some',
    ], alarms.alarm_on_segfault([alarm1, alarm2]))





from uwsgiconf import Section


def test_statics_basics(assert_lines):

    assert_lines([
        'static-index = index.html',
    ], Section().statics.set_basic_params(index_file='index.html'))

    assert_lines([
        'check-static-docroot = true',
    ], Section().statics.set_basic_params(static_dir=Section.statics.DIR_DOCUMENT_ROOT))

    assert_lines([
        'static-map2 = /static=/var/www/static',
        'static-safe = /var/www/static',

    ], Section().statics.register_static_map(
        '/static', '/var/www/static', retain_resource_path=True, safe_target=True
    ))

    assert_lines([
        'static-cache-paths = 200',
    ], Section().statics.set_paths_caching_params(timeout=200))


def test_statics_expiration(assert_lines):

    statics = Section().statics

    assert_lines([
        'static-expires-mtime = .*png 50',
        'static-expires-mtime = .*pdf 50',
        'static-expires-type = text/html=100',

    ],
    statics.add_expiration_rule(
        statics.expiration_criteria.FILENAME, ['.*png', '.*pdf'], 50, use_mod_time=True

    ).statics.add_expiration_rule(
        statics.expiration_criteria.MIME_TYPE, 'text/html', 100

    ))




import os
import io
import sys
from setuptools import setup, find_packages

from uwsgiconf import VERSION


PATH_BASE = os.path.dirname(__file__)
PYTEST_RUNNER = ['pytest-runner'] if 'test' in sys.argv else []


def get_readme():
    # This will return README (including those with Unicode symbols).
    with io.open(os.path.join(PATH_BASE, 'README.rst')) as f:
        return f.read()


setup(
    name='uwsgiconf',
    version='.'.join(map(str, VERSION)),
    url='https://github.com/idlesign/uwsgiconf',

    description='Configure uWSGI from your Python code',
    long_description=get_readme(),
    license='BSD 3-Clause License',

    author='Igor `idle sign` Starikov',
    author_email='idlesign@yandex.ru',

    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,

    install_requires=[],
    setup_requires=[] + PYTEST_RUNNER,

    entry_points={
        'console_scripts': ['uwsgiconf = uwsgiconf.cli:main'],
    },

    test_suite='tests',

    classifiers=[
        # As in https://pypi.python.org/pypi?:action=list_classifiers
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: BSD License'
    ],
)





import pytest

from uwsgiconf import Section
from uwsgiconf.exceptions import ConfigurationError


def test_routing_basics(assert_lines):

    assert_lines([
        'routers-list = true',
    ], Section().routing.print_routes())

    assert_lines([
        'error-page-500 = /here/500.html',
    ], Section().routing.set_error_page(500, '/here/500.html'))

    with pytest.raises(ConfigurationError):  # unsupported code
        assert_lines('', Section().routing.set_error_page(800, '/here/800.html'))


def test_routing_rules(assert_lines):
    # route-if = equal:${PATH_INFO};/bad break:500 Internal Server Error
    # response-route-if = equal:${MY_CONTENT_TYPE};text/html goto:gzipme
    # response-route-if = contains:${HTTP_ACCEPT_ENCODING};gzip gzip:
    # route-if = regexp:FOO;^F log:starts with F

    rule = Section.routing.route_rule

    assert_lines([
        'response-route-run = continue:',

    ], Section().routing.register_route(
        rule(
            rule.actions.do_continue(),
            subject=None,
            stage=rule.stages.RESPONSE)
    ))

    assert_lines([
        'route-if-not = eq:${PATH_INFO};/bad chdir:/here',

    ], Section().routing.register_route(
        rule(
            rule.actions.dir_change('/here'),
            ~rule.subjects.custom(rule.vars.request('PATH_INFO')).eq('/bad'))
    ))


def test_routing_subject_custom():

    subj = Section.routing.route_rule.subjects.custom('mysubj')

    assert str(subj.exists()) == 'exists:mysubj'
    assert str(subj.isfile()) == 'isfile:mysubj'
    assert str(subj.isdir()) == 'isdir:mysubj'
    assert str(subj.islink()) == 'islink:mysubj'
    assert str(subj.isexec()) == 'isexec:mysubj'

    assert str(subj.eq(33)) == 'eq:mysubj;33'
    assert str(subj.ge(33)) == 'ishigherequal:mysubj;33'
    assert str(subj.le(33)) == 'islowerequal:mysubj;33'
    assert str(subj.gt(33)) == 'ishigher:mysubj;33'
    assert str(subj.lt(33)) == 'islower:mysubj;33'

    assert str(subj.isempty()) == 'empty:mysubj'
    assert str(subj.startswith('my')) == 'startswith:mysubj;my'
    assert str(subj.endswith('bj')) == 'endswith:mysubj;bj'
    assert str(subj.matches('mys')) == 'regexp:mysubj;mys'
    assert str(subj.contains('sub')) == 'contains:mysubj;sub'
    assert str(subj.contains_ipv4()) == 'ipv4in:mysubj'
    assert str(subj.contains_ipv6()) == 'ipv6in:mysubj'

    assert str(subj.islord()) == 'lord:mysubj'


def test_routing_transformations(assert_lines):

    rule = Section.routing.route_rule

    assert_lines(
        'route-run = forcexcl:',
        Section().routing.register_route(rule(
            rule.transforms.fix_content_len(add_header=True), subject=None))
    )

    assert_lines(
        'route-run = flush:',
        Section().routing.register_route(rule(rule.transforms.flush(), subject=None))
    )

    assert_lines(
        'route-run = gzip:',
        Section().routing.register_route(rule(rule.transforms.gzip(), subject=None))
    )

    assert_lines(
        'route-run = toupper:',
        Section().routing.register_route(rule(rule.transforms.upper(), subject=None))
    )

    assert_lines(
        'route-run = tofile:filename=/tmp/mycache',
        Section().routing.register_route(rule(rule.transforms.to_file('/tmp/mycache'), subject=None))
    )

    assert_lines(
        'route-run = chunked:',
        Section().routing.register_route(rule(rule.transforms.chunked(), subject=None))
    )

    assert_lines(
        'route-run = template:',
        Section().routing.register_route(rule(rule.transforms.template(), subject=None))
    )


def test_routing_actions(assert_lines):

    rule = Section.routing.route_rule

    assert_lines([
        'route-run = continue:',

    ], Section().routing.register_route(
        rule(rule.actions.do_continue(), subject=None)
    ))

    assert_lines([
        'route-run = send-crnl:HTTP/1.0 100 Continue',

    ], Section().routing.register_route(
        rule(rule.actions.send('HTTP/1.0 100 Continue', crnl=True), subject=None)
    ))

    assert_lines([
        'plugin = router_uwsgi',
        'route-run = uwsgi:127.0.0.1:3031,5,,fooapp',

    ], Section().routing.register_route(
        rule(rule.actions.route_uwsgi('127.0.0.1:3031', mod1=5, app='fooapp'), subject=None)
    ))

    assert_lines([
        'plugin = router_rewrite',
        'route-uri = /some rewrite-last:/index.php?page=$1.php',

    ], Section().routing.register_route(
        rule(
            rule.actions.rewrite('/index.php?page=$1.php', do_continue=True),
            subject=rule.subjects.REQUEST_URI('/some'))
    ))

    assert_lines([
        'plugin = router_redirect',
        'route-run = redirect-301:http://here',

    ], Section().routing.register_route(
        rule(rule.actions.redirect('http://here', permanent=True), subject=None)
    ))

    assert_lines([
        'route-run = return:500',

    ], Section().routing.register_route(
        rule(rule.actions.do_break(500, return_body=True), subject=None)
    ))

    assert_lines([
        'route-run = donotlog:',

    ], Section().routing.register_route(
        rule(rule.actions.log(None), subject=None)
    ))

    assert_lines([
        'route-run = donotoffload:',

    ], Section().routing.register_route(
        rule(rule.actions.offload_off(), subject=None)
    ))

    assert_lines([
        'route-run = logvar:myvar $1',

    ], Section().routing.register_route(
        rule(rule.actions.add_var_log('myvar', '$1'), subject=None)
    ))

    assert_lines([
        'route-run = goto:here',

    ], Section().routing.register_route(
        rule(rule.actions.do_goto('here'), subject=None)
    ))

    assert_lines([
        'route-run = addvar:MY_CGI myval',

    ], Section().routing.register_route(
        rule(rule.actions.add_var_cgi('MY_CGI', 'myval'), subject=None)
    ))

    assert_lines([
        'route-run = addheader:Content-Type: text/html',

    ], Section().routing.register_route(
        rule(rule.actions.header_add('Content-Type', 'text/html'), subject=None)
    ))

    assert_lines([
        'route-run = delheader:Content-Type',

    ], Section().routing.register_route(
        rule(rule.actions.header_remove('Content-Type'), subject=None)
    ))

    assert_lines([
        'route-run = disableheaders:',

    ], Section().routing.register_route(
        rule(rule.actions.headers_off(), subject=None)
    ))

    assert_lines([
        'route-run = clearheaders:200',

    ], Section().routing.register_route(
        rule(rule.actions.headers_reset(200), subject=None)
    ))

    assert_lines([
        'route-run = clearheaders:200',

    ], Section().routing.register_route(
        rule(rule.actions.headers_reset(200), subject=None)
    ))

    assert_lines([
        'route-run = signal:33',

    ], Section().routing.register_route(
        rule(rule.actions.signal(33), subject=None)
    ))

    assert_lines([
        'plugin = router_http',
        'route-run = http:127.0.0.1:8181,my.domain.ru',

    ], Section().routing.register_route(
        rule(rule.actions.route_external('127.0.0.1:8181', 'my.domain.ru'), subject=None)
    ))


    assert_lines([
        'plugin = router_static',
        'route-run = static:/here/a.jpg',

    ], Section().routing.register_route(
        rule(rule.actions.serve_static('/here/a.jpg'), subject=None)
    ))

    assert_lines([
        'plugin = router_basicauth',
        'route-run = basicauth-next:my realm,usr:pwd',
        'route-run = basicauth:my realm,htpwd',
        'route-run = basicauth:my realm,ussr:',

    ], Section().routing.register_route([
        rule(rule.actions.auth_basic('my realm', 'usr', 'pwd', do_next=True), subject=None),
        rule(rule.actions.auth_basic('my realm', password='htpwd'), subject=None),
        rule(rule.actions.auth_basic('my realm', 'ussr'), subject=None),
    ]))

    assert_lines([
        'plugin = ldap',
        'route-run = ldapauth-next:my realm,url=ldap://ldap.domain.com;basedn=ou=users,dc=domain;loglevel=1',

    ], Section().routing.register_route([
        rule(rule.actions.auth_ldap(
            'my realm', 'ldap://ldap.domain.com', base_dn='ou=users,dc=domain', log_level=1,
            do_next=True), subject=None),
    ]))

    assert_lines([
        'route-run = chdir:/here',

    ], Section().routing.register_route([
        rule(rule.actions.dir_change('/here'), subject=None),
    ]))

    assert_lines([
        'route-run = setapp:myapp',

    ], Section().routing.register_route([
        rule(rule.actions.set_var_uwsgi_appid('myapp'), subject=None),
    ]))

    assert_lines([
        'route-run = setuser:iam',

    ], Section().routing.register_route([
        rule(rule.actions.set_var_remote_user('iam'), subject=None),
    ]))

    assert_lines(
        'route-run = harakiri:30',
        Section().routing.register_route(rule(rule.actions.set_harakiri(30), subject=None))
    )

    assert_lines(
        'route-run = sethome:/there',
        Section().routing.register_route(rule(rule.actions.set_var_uwsgi_home('/there'), subject=None))
    )

    assert_lines(
        'route-run = setscheme:uwsgi',
        Section().routing.register_route(rule(rule.actions.set_var_uwsgi_scheme('uwsgi'), subject=None))
    )

    assert_lines(
        'route-run = setscriptname:my.php',
        Section().routing.register_route(rule(rule.actions.set_var_script_name('my.php'), subject=None))
    )

    assert_lines(
        'route-run = setmethod:POST',
        Section().routing.register_route(rule(rule.actions.set_var_request_method('POST'), subject=None))
    )

    assert_lines(
        'route-run = seturi:http://there',
        Section().routing.register_route(rule(rule.actions.set_var_request_uri('http://there'), subject=None))
    )

    assert_lines(
        'route-run = setremoteaddr:127.0.0.1',
        Section().routing.register_route(rule(rule.actions.set_var_remote_addr('127.0.0.1'), subject=None))
    )

    assert_lines(
        'route-run = setpathinfo:/there/here.html',
        Section().routing.register_route(rule(rule.actions.set_var_path_info('/there/here.html'), subject=None))
    )

    assert_lines(
        'route-run = fixpathinfo:',
        Section().routing.register_route(
            rule(rule.actions.fix_var_path_info(), subject=None)
        )
    )

    assert_lines(
        'route-run = setfile:/var/uwsgi/app002.py',
        Section().routing.register_route(rule(rule.actions.set_script_file('/var/uwsgi/app002.py'), subject=None))
    )

    assert_lines(
        'route-run = setdocroot:/var/here',
        Section().routing.register_route(rule(rule.actions.set_var_document_root('/var/here'), subject=None))
    )

    assert_lines(
        'route-run = setprocname:myproc',
        Section().routing.register_route(rule(rule.actions.set_uwsgi_process_name('myproc'), subject=None))
    )

    uwsgi = rule.vars.uwsgi
    assert_lines(
        'route-run = alarm:pippo ${uwsgi[wid]} ${uwsgi[pid]}',
        Section().routing.register_route(rule(
            rule.actions.alarm('pippo', '%s %s' % (uwsgi('wid'), uwsgi('pid'))), subject=None
        ))
    )


def test_routing_goto_label(assert_lines):

    rule = Section.routing.route_rule
    actions = rule.actions
    subjects = rule.subjects

    label = 'localhost'

    assert_lines([
            'route-host = ^localhost$ goto:localhost',

            'route-label = localhost',
            'plugin = router_redirect',
            'route-user-agent = .*curl.* redirect-302:http://uwsgi.it',
            'route = (.*) continue:',
        ],

        Section().routing.register_route(

            rule(
                actions.do_goto(label),
                subjects.HTTP_HOST('^localhost$')
            )

        ).routing.register_route(
            [
                rule(
                    actions.redirect('http://uwsgi.it'),
                    subjects.HTTP_USER_AGENT('.*curl.*')
                ),
                rule(
                    actions.do_continue()
                ),
            ], label=label)
    )



import os
from copy import deepcopy
from datetime import datetime
from collections import OrderedDict
from functools import partial

from .base import Options, OptionsGroup
from .options import *
from .formatters import IniFormatter, format_print_text
from .exceptions import ConfigurationError
from .utils import listify


class Section(OptionsGroup):
    """Configuration section.

    Options within configuration section are gathered into groups:

        * alarms
        * caching
        * master_process
        * workers
        * etc.

    Next to all public methods of groups are for setting configuration parameters.
    Such methods return section object to allow chaining.

    You can pass options group basic parameters into (the following are all the same):

        * ``set_basic_params()`` as in ``section.workers.set_basic_params(count=3)``

        * ``__call__`` as in ``section.workers(count=3)``

        * section initializer using `params_` prefixed group name:

            .. code-block:: python
                Section(
                    params_workers=dict(count=3),
                )

    """
    alarms = Options(Alarms)  # type: Alarms
    """Alarms options group."""

    applications = Options(Applications)  # type: Applications
    """Applications options group."""

    caching = Options(Caching)  # type: Caching
    """Caching options group."""

    cheapening = Options(Cheapening)  # type: Cheapening
    """Cheapening options group."""

    empire = Options(Empire)  # type: Empire
    """Emperor and vassals options group."""

    locks = Options(Locks)  # type: Locks
    """Locks options group."""

    logging = Options(Logging)  # type: Logging
    """Logging options group."""

    main_process = Options(MainProcess)  # type: MainProcess
    """Main process options group."""

    master_process = Options(MasterProcess)  # type: MasterProcess
    """Master process options group."""

    monitoring = Options(Monitoring)  # type: Monitoring
    """Monitoring options group."""

    networking = Options(Networking)  # type: Networking
    """Networking options group."""

    queue = Options(Queue)  # type: Queue
    """Queue options group."""

    routing = Options(Routing)  # type: Routing
    """Routing related options group."""

    spooler = Options(Spooler)  # type: Spooler
    """Spooler options group."""

    statics = Options(Statics)  # type: Statics
    """Static file serving options group."""

    workers = Options(Workers)  # type: Workers
    """Workers options group."""

    python = Options(Python)  # type: Python
    """Python options group."""

    def __init__(self, name=None, strict_config=None, style_prints=False, **kwargs):
        """

        :param bool strict_config: Enable strict configuration parsing.
            If any unknown option is encountered in a configuration file,
            an error is shown and uWSGI quits.

            To use placeholder variables when using strict mode, use the ``set-placeholder`` option.

        :param str name: Configuration section name.

        :param bool style_prints: Enables styling (e.g. colouring) for ``print_`` family methods.
            Could be nice for console and distracting in logs.

        """
        self._style_prints = style_prints
        self._plugins = []

        self._section = self
        self._options_objects = OrderedDict()
        self._opts = OrderedDict()

        self.name = name or 'uwsgi'

        super(Section, self).__init__(**kwargs)

        self._set_basic_params_from_dict(kwargs)
        self.set_basic_params(strict_config=strict_config)

    def set_basic_params(self, strict_config=None, **kwargs):

        self._set('strict', strict_config, cast=bool)

        return self

    def as_configuration(self):
        """Returns configuration object including only one (this very) section.

        :rtype: Configuration
        """
        return Configuration([self])

    def print_plugins(self):
        """Print out enabled plugins."""

        self._set('plugins-list', True, cast=bool)

        return self

    def print_stamp(self):
        """Prints out a stamp containing useful information,
        such as what and when has generated this configuration.

        """
        from . import VERSION

        print_out = partial(self.print_out, format_options='red')
        print_out('This configuration was automatically generated using')
        print_out('uwsgiconf v%s on %s' % ('.'.join(map(str, VERSION)), datetime.now().isoformat(' ')))

        return self

    def print_out(self, value, indent=None, format_options=None, asap=False):
        """Prints out the given value.

        :param value:

        :param str|unicode indent:

        :param dict|str|unicode format_options: text color

        :param bool asap: Print as soon as possible.

        """
        if indent is None:
            indent = '>   '

        text = indent + str(value)

        if format_options is None:
            format_options = 'gray'

        if self._style_prints and format_options:

            if not isinstance(format_options, dict):
                format_options = {'color_fg': format_options}

            text = format_print_text(text, **format_options)

        command = 'iprint' if asap else 'print'
        self._set(command, text, multi=True)

        return self

    def print_variables(self):
        """Prints out magic variables available in config files
        alongside with their values and descriptions.
        May be useful for debugging.

        http://uwsgi-docs.readthedocs.io/en/latest/Configuration.html#magic-variables

        """
        print_out = partial(self.print_out, format_options='green')

        print_out('===== variables =====')

        for var, hint in self.vars.get_descriptions().items():
            print_out('    %' + var + ' = ' + var + ' = ' + hint.replace('%', '%%'))

        print_out('=====================')

        return self

    def set_plugins_params(self, plugins=None, search_dirs=None, autoload=None, required=False):
        """Sets plugin-related parameters.

        :param list|str|unicode|OptionsGroup|list[OptionsGroup] plugins: uWSGI plugins to load

        :param list|str|unicode search_dirs: Directories to search for uWSGI plugins.

        :param bool autoload: Try to automatically load plugins when unknown options are found.

        :param bool required: Load uWSGI plugins and exit on error.

        """
        plugins = plugins or []

        command = 'need-plugin' if required else 'plugin'

        for plugin in listify(plugins):

            if plugin not in self._plugins:
                self._set(command, plugin, multi=True)
                self._plugins.append(plugin)

        self._set('plugins-dir', search_dirs, multi=True)
        self._set('autoload', autoload, cast=bool)

        return self

    def set_fallback(self, target):
        """Sets a fallback configuration for section.

        Re-exec uWSGI with the specified config when exit code is 1.

        :param str|unicode|Section target: File path or Section to include.
        """
        if isinstance(target, Section):
            target = ':' + target.name

        self._set('fallback-config', target)

        return self

    def env(self, key, value=None, unset=False, asap=False):
        """Processes (sets/unsets) environment variable.

        If is not given in `set` mode value will be taken from current env.

        :param str|unicode key:

        :param value:

        :param bool unset: Whether to unset this variable.

        :param bool asap: If True env variable will be set as soon as possible.

        """
        if unset:
            self._set('unenv', key, multi=True)
        else:
            if value is None:
                value = os.environ.get(key)

            self._set('%senv' % ('i' if asap else ''), '%s=%s' % (key, value), multi=True)

        return self

    def include(self, target):
        """Includes target contents into config.

        :param str|unicode|Section|list target: File path or Section to include.
        """
        for target_ in listify(target):
            if isinstance(target_, Section):
                target_ = ':' + target_.name

            self._set('ini', target_, multi=True)

        return self

    @classmethod
    def derive_from(cls, section, name=None):
        """Creates a new section based on the given.

        :param Section section: Section to derive from,

        :param str|unicode name: New section name.

        :rtype: Section
        """
        new_section = deepcopy(section)

        if name:
            new_section.name = name

        return new_section

    def _set_basic_params_from_dict(self, src_dict):

        for key, value in src_dict.items():
            if not key.startswith('params_') or not value:
                continue

            group_attr_name = key.replace('params_', '')
            options_group = getattr(self, group_attr_name, None)  # type: OptionsGroup

            if options_group is not None:
                options_group.set_basic_params(**value)

    def _get_options(self):
        options = []

        for name, val in self._section._opts.items():

            for val_ in listify(val):
                options.append((name, val_))

        return options

    class vars(object):
        """The following variables also known as magic variables
        could be used as option values where appropriate.

        * http://uwsgi-docs.readthedocs.io/en/latest/Configuration.html#magic-variables

        """
        VERSION = '%V'
        """uWSGI version number"""

        FORMAT_ESCAPE = '%['
        """ANSI escape \\033. useful for printing colors"""

        FORMAT_END = '%s[0m' % FORMAT_ESCAPE

        CONF_CURRENT_SECTION = '%x'
        """The current section identifier, eg. conf.ini:section."""
        CONF_CURRENT_SECTION_NTPL = '%x'

        CONF_NAME_ORIGINAL = '%o'
        """The original conf filename, as specified on the command line"""
        CONF_NAME_ORIGINAL_NTPL = '%O'

        TIMESTAMP_STARTUP_S = '%t'
        """Unix time s, gathered at instance startup."""
        TIMESTAMP_STARTUP_MS = '%T'
        """Unix time ms, gathered at instance startup"""

        DIR_VASSALS = '%v'
        """Vassals directory - pwd."""

        HOST_NAME = '%h'
        """Host name."""
        CPU_CORES = '%k'
        """Detected CPU count."""

        USER_ID = '%u'
        """User ID."""
        USER_NAME = '%U'
        """User name."""

        GROUP_ID = '%g'
        """Use group ID."""
        GROUP_NAME = '%G'
        """Use group name."""

        @classmethod
        def get_descriptions(cls):
            """Returns variable to description mapping.

            :rtype: dict
            """
            descriptions = {
                cls.DIR_VASSALS: 'the vassals directory - pwd',
                cls.VERSION: 'the uWSGI version',
                cls.HOST_NAME: 'the hostname',
                cls.CONF_NAME_ORIGINAL: 'the original conf filename, as specified on the command line',
                cls.CONF_NAME_ORIGINAL_NTPL: 'as %o but for first non-template conf',
                '%p': 'the absolute path of the conf',
                '%P': 'as %p but for first non-template conf',
                '%s': 'the filename of the conf',
                '%S': 'as %s but for first non-template conf',
                '%d': 'the absolute path of the directory containing the conf',
                '%D': 'as %d but for first non-template conf',
                '%e': 'the extension of the conf',
                '%E': 'as %e but for first non-template conf',
                '%n': 'the filename without extension',
                '%N': 'as %n but for first non-template conf',
                '%c': 'the name of the directory containing the conf file',
                '%C': 'as %c but for first non-template conf',
                cls.TIMESTAMP_STARTUP_S: 'unix time s, gathered at instance startup',
                cls.TIMESTAMP_STARTUP_MS: 'unix time ms, gathered at instance startup',
                cls.CONF_CURRENT_SECTION: 'the current section identifier, eg. conf.ini:section',
                '%X': 'as %x but for first non-template conf',
                '%i': 'inode number of the file',
                '%I': 'as %i but for first non-template conf',
                cls.FORMAT_ESCAPE: 'ANSI escape \\033. useful for printing colors',
                cls.CPU_CORES: 'detected cpu cores',
                cls.USER_ID: 'uid of the user',
                cls.USER_NAME: 'username or fallback to uid of the user',
                cls.GROUP_ID: 'gid of the user',
                cls.GROUP_NAME: 'group name or fallback to gid of the user',
                '%j': 'HEX representation of the djb33x hash of the full conf path',
                '%J': 'as %j but for first non-template conf',
            }

            descriptions = sorted(descriptions.items(), key=lambda item: item[0].lower())

            return OrderedDict(descriptions)


class Configuration(object):
    """
    Configuration is comprised from one or more Sections and could
    be represented in format natively supported by uWSGI.

    """

    def __init__(self, sections=None, autoinclude_sections=False):
        """

        :param list[Section] sections: If not provided, empty section
            will be automatically generated.

        :param bool autoinclude_sections: Whether to include
            in the first sections all subsequent sections.

        """
        super(Configuration, self).__init__()

        sections = sections or [Section()]
        self._validate_sections(sections)

        if autoinclude_sections:

            first = sections[0]
            for section in sections[1:]:
                first.include(section)

        self.sections = sections

    @classmethod
    def _validate_sections(cls, sections):
        """Validates sections types and uniqueness."""
        names = []
        for section in sections:

            if not hasattr(section, 'name'):
                raise ConfigurationError('`sections` attribute requires a list of Section')

            name = section.name
            if name in names:
                raise ConfigurationError('`%s` section name must be unique' % name)

            names.append(name)

    def format(self, do_print=False, stamp=True):
        """Applies formatting to configuration.

        *Currently formats to .ini*

        :param bool do_print: Whether to print out formatted config.
        :param bool stamp: Whether to add stamp data to the first configuration section.
        :rtype: str|unicode
        """
        if stamp and self.sections:
            self.sections[0].print_stamp()

        formatted = IniFormatter(self.sections).format()

        if do_print:
            print(formatted)

        return formatted

    def print_ini(self):
        """Print out this configuration as .ini.

        :rtype: str|unicode
        """
        return self.format(do_print=True)



from collections import OrderedDict
from .utils import listify


if False:  # pragma: nocover
    from .config import Section


class Options(object):
    """Options descriptor. Allows option."""

    def __init__(self, opt_type):
        """
        :param opt_type:
        """
        self.opt_type = opt_type

    def __get__(self, section, section_cls):
        """

        :param Section section:
        :param OptionsGroup options_obj:
        :rtype: OptionsGroup
        """
        key = self.opt_type.__name__

        try:
            options_obj = section._options_objects.get(key)

        except AttributeError:
            # Allow easy access to option group static params:
            # Section.networking.socket_types.DEFAULT
            return self.opt_type

        if not options_obj:
            options_obj = self.opt_type(_section=section)

        section._options_objects[key] = options_obj

        return options_obj


class OptionsGroup(object):
    """Introduces group of options.

    Basic group parameters may be passed to initializer
    or `set_basic_params` method.

    Usually methods setting parameters are named `set_***_params`.

    Methods ending with `_params` return section object and may be chained.

    """
    _section = None  # type: Section
    """Section this option group belongs to."""

    plugin = False  # type: bool|str|unicode
    """Indication this option group belongs to a plugin."""

    name = None
    """Name to represent the group."""

    def __init__(self, *args, **kwargs):
        if self._section is None:
            self._section = kwargs.pop('_section', None)  # type: Section

        self.set_basic_params(*args, **kwargs)

    def _get_name(self, *args, **kwargs):
        """
        :rtype: str
        """
        return self.name

    def __str__(self):
        return self._get_name()

    def __call__(self, *args, **kwargs):
        """The call is translated into ``set_basic_params``` call.

        This approach is much more convenient yet IDE most probably won't
        give you a hint on what arguments are accepted.

        :param args:
        :param kwargs:
        :rtype: Section
        """
        return self.set_basic_params(*args, **kwargs)

    def set_basic_params(self, *args, **kwargs):
        """
        :rtype: Section
        """
        return self._section

    def _set(self, key, value, condition=True, cast=None, multi=False, plugin=None):

        def set_plugin(plugin):
            self._section.set_plugins_params(plugins=plugin)

        def handle_plugin_required(val):

            if isinstance(val, ParametrizedValue):
                if val.plugin:
                    # Automatic plugin activation.
                    set_plugin(val.plugin)

                if val._opts:
                    opts.update(val._opts)

        if condition is True:
            condition = value is not None

        if condition is None or condition:

            opts = self._section._opts

            if cast is bool:
                if value:
                    value = 'true'

                else:
                    try:
                        del opts[key]
                    except KeyError:
                        pass

                    return

            if self.plugin is True:
                # Automatic plugin activation when option from it is used.
                set_plugin(self)

            if plugin:
                set_plugin(plugin)

            if isinstance(value, tuple):  # Tuple - expect ParametrizedValue.
                list(map(handle_plugin_required, value))
                value = ' '.join(map(str, value))

            if multi:
                values = []

                # First activate plugin if required.
                for value in listify(value):
                    handle_plugin_required(value)
                    values.append(value)

                # Second: list in new option.
                opts.setdefault(key, []).extend(values)

            else:
                handle_plugin_required(value)
                opts[key] = value


class ParametrizedValue(OptionsGroup):
    """Represents parametrized option value."""

    alias = None
    args_joiner = ' '
    name_separator = ':'
    name_separator_strip = False

    def __init__(self, *args):
        self.args = args
        self._opts = OrderedDict()
        super(ParametrizedValue, self).__init__(_section=self)

    def __str__(self):
        args = [str(arg) for arg in self.args if arg is not None]

        result = self._get_name() + self.name_separator

        result += self.args_joiner.join(args)

        if self.alias:
            result = '%s %s' % (self.alias, result)

        result = result.strip()

        if self.name_separator_strip:
            result = result.strip(self.name_separator)

        return result



from uwsgiconf import Section


def test_applications_basics(assert_lines):

    assert_lines([
        'need-app = true',
    ], Section().applications.set_basic_params(exit_if_none=True))

    assert_lines([
        'chdir = /here',
    ], Section().applications.change_dir('/here'))

    assert_lines([
        'mount = /articles=app.py',
    ], Section().applications.mount('/articles', 'app.py'))

    assert_lines([
        'lazy-apps = true',
    ], Section().applications.switch_into_lazy_mode())
